import mathutils
import bge
import math

def init(controller):
  obj = controller.owner
  scene = bge.logic.getCurrentScene()
  # Make the background black even when using the Embedded Player
  bge.render.setBackgroundColor(mathutils.Vector((0, 0, 0, 1)))

  def grid_snap_floor(a):
    return math.floor(a / spacing) * spacing

  obj["grid_snap_floor"] = grid_snap_floor

  def draw_map(startx, starty, old_startx, old_starty, width, height):
    for i in range(startx, startx + width, spacing): 
      for j in range(starty, starty + height, spacing): 
        if ((i < old_startx or i > old_startx + width - spacing or
            j < old_starty or j > old_starty + height - spacing) and
            ((i not in obj["eaten"]) or (j not in obj["eaten"][i]))):
          # Must use an unscaled object for the second argument. Otherwise the
          # first object's scale will be multiplied by the second object's scale
          new_cube = scene.addObject("Cube prototype", obj)
          new_cube.worldPosition = mathutils.Vector((i, j, 0))

  obj["draw_map"] = draw_map

  spacing = obj["spacing"]

  obj["eaten"] = {}

  sdim = mathutils.Vector()
  map_edge = mathutils.Vector()
  map_dim = mathutils.Vector()

  camera = scene.active_camera

  # Tip: use the projection_matrix rather than getScreenVect().
  # getScreenVect() is buggy
  # Some reading for projection matrices:
  # http://www.songho.ca/opengl/gl_projectionmatrix.html
  # http://robertokoci.com/world-view-projection-matrix-unveiled/
  # http://http.developer.nvidia.com/CgTutorial/cg_tutorial_chapter04.html

  inverse_projection = camera.projection_matrix.inverted() * camera.modelview_matrix.inverted()
  bot_right = mathutils.Vector((1, -1, 0)) * inverse_projection
  top_left = mathutils.Vector((-1, 1, 0)) * inverse_projection

  sdim.x = bot_right.x - top_left.x
  sdim.y = top_left.y - bot_right.y

  map_edge.x = grid_snap_floor(top_left.x) - spacing
  map_edge.y = grid_snap_floor(bot_right.y) - spacing

  map_dim.x = grid_snap_floor(bot_right.x) + spacing * 2 - map_edge.x
  map_dim.y = grid_snap_floor(top_left.y) + spacing * 2 - map_edge.y

  draw_map(int(map_edge.x), int(map_edge.y),
    int(map_edge.x - map_dim.x), int(map_edge.y - map_dim.y),
    int(map_dim.x), int(map_dim.y))

  # Fortunately these seem to work. Otherwise I'd be stuck using
  # bge.logic.globalDict, or /shudder/ a String property of the object
  obj["sdim"] = sdim
  obj["map_edge"] = map_edge
  obj["map_dim"] = map_dim
  obj["camera"] = camera

def update(controller):
  obj = controller.owner

  if "map_dim" in obj:
    spacing = obj["spacing"]
    sdim = obj["sdim"]
    map_edge = obj["map_edge"]
    map_dim = obj["map_dim"]
    camera = obj["camera"]

    # Storing functions in the dictionary feels weird, but I still like it
    # better than using module-level variables or passing spacing, eaten, 
    # and scene as arguments on every call
    draw_map = obj["draw_map"]
    grid_snap_floor = obj["grid_snap_floor"]

    old_edge = mathutils.Vector()
    cpos = camera.worldPosition - sdim / 2

    if ((cpos.y < map_edge.y + spacing) or
       (cpos.y + sdim.y > map_edge.y + map_dim.y - spacing) or
       (cpos.x < map_edge.x + spacing) or
       (cpos.x + sdim.x > map_edge.x + map_dim.x - spacing)):

      old_edge.x = map_edge.x
      old_edge.y = map_edge.y
      map_edge.x = grid_snap_floor(cpos.x) - spacing
      map_edge.y = grid_snap_floor(cpos.y) - spacing
      obj["map_edge"] = map_edge

      draw_map(int(map_edge.x), int(map_edge.y),
        int(old_edge.x), int(old_edge.y),
        int(map_dim.x), int(map_dim.y))
